#########################################################################
#                                                                       #
# Script:  notebook.tcl                                                 #
#                                                                       #
# Author:  Dave Freese, dave@darkmatterfactory.com                      #
#                                                                       #
# Provides:  notebook <widget path> [options]                           #
#                                                                       #
# Purpose:  Provides a mega-widget using tabbed notebook metaphor       #
#                                                                       #
# Credits:                                                              #
#           D. Richard Hipp                                             #
#           I borrowed freely and significantly modified his notebook   #
#           widget.  The design of the widget and its creation is the   #
#           same as 'calendar', 'dateentry' and 'combobox' all of       #
#           which use a megawidget design originated by Bryan Oakley    #
#           in his original combobox widget. (oakley@channelpoint.com)  #
#                                                                       #
# Exit Codes:  0  Ok                                                    #
#                                                                       #
# Version:  v0.1.0                                                      #
# --------------------------------------------------------------------- #
#                                                                       #
#                                                                       #
# Version:  0.1.0                                                       #
# Date:    9 October, 2001                                              #
# Author:  Dave Freese, dave@darkmatterfactory.com                      #
# Change:  Original released version.                                   #
#                                                                       #
# Version:                                                              #
# Date:                                                                 #
# Author:                                                               #
# Change:                                                               #
#                                                                       #
#########################################################################

###############################################################################
#    notebook.tcl - tcl module to provide calendar mega-wdget                 #
#    Copyright (C) 2001 Dave Freese                                           #
#                                                                             #
#    This program is free software; you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation; either version 2 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program; if not, write to the Free Software              #
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA  #
#                                                                             #
###############################################################################

package require Tk 8.0
package provide Notebook 0.1

package require Megawidget 0.1

namespace eval ::notebook {

# this is the public interface
  namespace export notebook

# these contain references to available options
  variable widgetOptions

# these contain references to available commands and subcommands
  variable widgetCommands

}

# ::notebook::notebook --
#
#     This is the command that gets exported. It creates a new
#     notebook widget.
#
# Arguments:
#
#     w        path of new widget to create
#     args     additional option/value pairs (eg: -background white, etc.)
#
# Results:
#
#     It creates the widget and sets up all of the default bindings
#
# Returns:
#
#     The name of the newly create widget

proc ::notebook::notebook {w args} {
  variable widgetOptions
  variable widgetCommands

  if {![info exists widgetOptions]} {
    Init
  }
  eval Build $w $args

  return $w
}


# ::notebook::Init --
#
#     Initialize the namespace variables. This should only be called once,
#     immediately prior to creating the first instance of the widget
#
# Arguments:
#
#    none
#
# Results:
#
#     All state variables are set to their default values; all of 
#     the option database entries will exist.
#
# Returns:
# 
#     empty string

proc ::notebook::Init {} {
  variable widgetOptions
  variable widgetCommands

  array set widgetOptions [list \
    -background          {background          Background} \
    -bg                  -background \
    -pageborderwidth     {pageBorderWidth     PageBorderWidth} \
    -command             {command             Command} \
    -commandstate        {commandState        State} \
    -cursor              {cursor              Cursor} \
    -fg                  -foreground \
    -foreground          {foreground          Foreground} \
    -disabledforeground  {disabledForeground  DisabledForeground} \
    -height              {height              Height} \
    -pad                 {pad                 Pad} \
    -pages               {pages               Pages} \
    -pagebordercolor     {pageBorderColor     PageBorderColor} \
    -pageborderhilight   {pageBorderHilight   PageBorderHilight} \
    -pagecolor           {pageColor           PageColor} \
    -pageheight          {pageHeight          PageHeight} \
    -pagewidth           {pageWidth           PageWidth} \
    -state               {state               State} \
    -top                 {top                 Top} \
    -width               {width               Width}
  ]


  set widgetCommands [list cget configure \
                           raise nbframe top \
                           expand scheduleExpand]

  option add *Notebook.top                 0
  option add *Notebook.pad                 2
  option add *Notebook.pages               {}
  option add *Notebook.background          grey80
  option add *Notebook.foreground          black
  option add *Notebook.disabledForeground  grey50
  option add *Notebook.pageColor           grey80
  option add *Notebook.pageWidth           200
  option add *Notebook.pageHeight          200
  option add *Notebook.pageBorderColor     grey20
  option add *Notebook.pageBorderHilight   white
  option add *Notebook.borderWidth         2

# set class bindings
  SetClassBindings
}

# ::notebook::SetClassBindings --
#
#    Sets up the default bindings for the widget class
#
#    this proc exists since it's The Right Thing To Do, but
#    I haven't had the time to figure out how to do all the
#    binding stuff on a class level. The main problem is that
#    the entry widget must have focus for the insertion cursor
#    to be visible. So, I either have to have the entry widget
#    have the Notebook bindtag, or do some fancy juggling of
#    events or some such. What a pain.
#
# Arguments:
#
#    none
#
# Returns:
#
#    empty string

proc ::notebook::SetClassBindings {} {

# make sure we clean up after ourselves...
  bind Notebook <Destroy> [list ::megawidget::DestroyHandler %W]
  return ""
}

# ::notebook::Build --
#
#    This does all of the work necessary to create the basic
#    notebook. 
#
# Arguments:
#
#    w        widget name
#    args     additional option/value pairs
#
# Results:
#
#    Creates a new widget with the given name. Also creates a new
#    namespace patterened after the widget name, as a child namespace
#    to ::notebook
#
# Returns:
#
#    the name of the widget

proc ::notebook::Build {w args } {
  variable widgetOptions

  if {[winfo exists $w]} {
    error "window name \"$w\" already exists"
  }

  namespace eval ::notebook::$w {

    variable options
    variable this
    variable pages

    set widgets(foo) foo  ;# coerce into an array
    set options(foo) foo  ;# coerce into an array

    unset widgets(foo)
    unset options(foo)
  }

  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options

  set widgets(this)   [frame $w -class Notebook \
                       -bd 0  \
                       -highlightthickness 0 \
                       -relief flat\
                       -takefocus 0]

  foreach name [array names widgetOptions] {
    if {[llength $widgetOptions($name)] == 1} continue
    set optName  [lindex $widgetOptions($name) 0]
    set optClass [lindex $widgetOptions($name) 1]
    set value [option get $w $optName $optClass]
    set options($name) $value
  }

  set widgets(canvas) [canvas $widgets(this).c \
    -bd 0 \
    -highlightthickness 0 \
    -takefocus 0 ]

  set thiswidget {[winfo parent %W]}
  bind $widgets(canvas) <1>   "::notebook::Click $thiswidget %x %y"

  pack $widgets(canvas) -side top
  
  set widgets(notebk) ::notebook::${w}::$w

#  array set options $args

  rename ::$w $widgets(notebk)

# Create our widget proc. All notebook widgets share the same widget proc.

  proc ::$w {command args} \
    "eval ::notebook::WidgetProc $w \$command \$args"

  if {[catch "::notebook::Configure $widgets(this) [array get options]" error]} {
    catch {destroy $w}
    error $error
  }

  if {[info exists options(-value)] \
      && [string length $options(-value)] == 0} {
    unset options(-value)
  }

# Now take care of configuration items passed on the arguement list...

  if { [llength $args] > 0 } {
    if { [catch "::notebook::Configure $widgets(this) $args" error] } {
      catch {destroy $w}
      error $error
    }
  }

  drawNotebook $w

  return ""
}

# ::notebook::WidgetProc --
#
#    This gets used as the widgetproc for an instance of a notebook 
#    widget.  Notice where the widget is created and you'll see that
#    the actual widget proc merely evals this proc with all of the
#    arguments intact.
#
#    Note that some widget commands are defined "inline" (ie:
#    within this proc), and some do most of their work in 
#    separate procs. This is merely because sometimes it was
#    easier to do it one way or the other.
#
# Arguments:
#
#    w         widget pathname
#    command   widget subcommand
#    args      additional arguments; varies with the subcommand
#
# Results:
#
#    Performs the requested widget command

proc ::notebook::WidgetProc {w command args} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options

  set command [::megawidget::Canonize $w command $command]

  set result ""

  switch $command {
    cget {
      if {[llength $args] != 1} {
        error "wrong # args: should be $w cget option"
      }
      set opt [::megawidget::Canonize $w option [lindex $args 0]]
      if {$opt == "-value"} {
        set result [$widget(entry) get]
      } else {
        set result $options($opt)
      }
    }
    
    nbframe {
      set result [eval ::notebook::Nbframe ${w} [lindex $args 0]]
    }
    
    raise {
      set result [eval ::notebook::Raise ${w} [lindex $args 0]]
    }
    
    expand {
      set result [eval ::notebook::Expand ${w} [lindex $args 0]]
    }
    
    scheduleExpand {
      set result [eval ::notebook::ScheduleExpand ${w} [lindex $args 0]]
    }
    
    top {
      set result [eval ::notebook::Top ${w}]
    }
    
    configure {
      set result [eval ::notebook::Configure ${w} $args]
      if {$result == 0} {
        drawNotebook $w
      }
    }
    default {
      error "bad option \"$command\""
    }
  }
  return $result
}

# ::notebook::Configure --
#
#    Implements the "configure" widget subcommand
#
# Arguments:
#
#    w      widget pathname
#    args   zero or more option/value pairs (or a single option)
#
# Results:
#    
#    Performs typcial "configure" type requests on the widget

proc ::notebook::Configure {w args} {
  variable widgetOptions
  variable defaultEntryCursor

  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options

# No arg's - return all of the configuration info
# config alias' has a length of 1
  if {[llength $args] == 0} {
    set results {}
    foreach opt [lsort [array names widgetOptions]] {
      if {[llength $widgetOptions($opt)] == 1} {
        set alias $widgetOptions($opt)
        set optName $widgetOptions($alias)
        lappend results [list $opt $optName]
      } else {
        set optName  [lindex $widgetOptions($opt) 0]
        set optClass [lindex $widgetOptions($opt) 1]
        set default [option get $w $optName $optClass]
        lappend results [list $opt $optName $optClass \
          $default $options($opt)]
      }
    }
    return $results
  }
    
# One argument - return configuration information on a single option
  if {[llength $args] == 1} {
    set opt [::megawidget::Canonize $w option [lindex $args 0]]
    set optName  [lindex $widgetOptions($opt) 0]
    set optClass [lindex $widgetOptions($opt) 1]
    set default [option get $w $optName $optClass]
    set results [list $opt $optName $optClass \
          $default $options($opt)]
    return $results
  }

# Odd number of values - return error message. 
  if {[expr {[llength $args]%2}] == 1} {
    error "value for \"[lindex $args end]\" missing"
  }
    
# Canonize returns the canonical option name or raises an error
# if the options name is not found.
  foreach {name value} $args {
    set name [::megawidget::Canonize $w option $name]
    set opts($name) $value
  }

# process all of the configuration options
  foreach option [array names opts] {
    if {[info exists options($option)]} {
      if {$option == "-width" || $option == "-height"} {
        continue
      }
      set options($option) $opts($option)
    }
  }  
  return 0
}

proc ::notebook::drawNotebook {w} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options

  $widgets(canvas) delete all

#specify tab dimensions based on text size
  if {![info exists options(-textheight)]} {
    set id [$widgets(canvas) create text 0 0 -text "T"]
    set bbox [$widgets(canvas) bbox $id]
    set options(-textheight) [expr [lindex $bbox 3] - [lindex $bbox 1] - 2]
    $widgets(canvas) delete $id
    foreach p $options(-pages) {
      set id [$widgets(canvas) create text 0 0 -text $p]
      set bbox [$widgets(canvas) bbox $id]
      set options($p,width) [expr [lindex $bbox 2] - [lindex $bbox 0] - 2]
      $widgets(canvas) delete $id
    }
# distance from top, bottom or left of tab to text is 1/3 text height
    set options(-tabpad_y) [expr $options(-textheight) / 3]
    set options(-tabpad_x) $options(-tabpad_y)
    set options(-tabh) [expr 2 * $options(-tabpad_y) + $options(-textheight)]
  }
    
  set options(x1) [expr $options(-pad) + $options(-pageborderwidth)]
  set options(x2) [expr $options(x1)   + $options(-pageborderwidth)]
  set options(x3) [expr $options(x2)   + $options(-pagewidth)]
  set options(x4) [expr $options(x3)   + $options(-pageborderwidth)]

  set options(-width) [expr $options(x4) + $options(-pad)]
  
  set options(x5) [expr 2 * $options(-tabpad_x)]
  set options(x6) [expr $options(x5)   + $options(-pageborderwidth)]
  set options(x7) [expr $options(x6)   + $options(-pageborderwidth)]
  
  set options(y1) [expr $options(-pad) + 2 * $options(-pageborderwidth)]
  set options(y2) [expr $options(y1)   + $options(-pageborderwidth)]
  set options(y5) [expr $options(y1)   + $options(-tabh)]
  set options(y6) [expr $options(y5)   + $options(-pageborderwidth)]
  set options(y3) [expr $options(y6)   + $options(-pageheight)]
  set options(y4) [expr $options(y3)   + $options(-pageborderwidth)]

  set options(-height) [expr $options(y4) + $options(-pad)]

  set y7          [expr $options(y1)   + $options(-tabpad_y)]

  
# draw the filled polygon for the body of the page
  $widgets(canvas) create poly \
    $options(x1) $options(y3) \
    $options(x2) $options(y4) \
    $options(x3) $options(y4) \
    $options(x4) $options(y3) \
    $options(x4) $options(y6) \
    $options(x3) $options(y5) \
    $options(x1) $options(y5) \
    -fill $options(-pagecolor)

  set x $options(x1)
  set cnt 0

  foreach p $options(-pages) {
    set options(p$cnt,x5) $x
    set width $options($p,width)
    
# draw the filled polygon for the tab
    $widgets(canvas) create poly \
       $x                                     $options(y6) \
       $x                                     $options(y2) \
       [expr $x + $options(-pageborderwidth)] $options(y1) \
       [expr $x + $options(x5) + $width]      $options(y1) \
       [expr $x + $options(x6) + $width]      $options(y2) \
       [expr $x + $options(x6) + $width]      $options(y6) \
       -fill $options(-pagecolor) -tags p$cnt

    set id [$widgets(canvas) create text 0 0 \
              -text $p -anchor nw -tags "p$cnt t$cnt"]
    $widgets(canvas) move $id [expr $x + $options(-tabpad_x)] $y7
    
# draw highlight border around left and top sides of tab
    $widgets(canvas) create line \
       $x $options(y5)\
       $x $options(y2) \
       [expr $x + $options(-pageborderwidth)]  $options(y1) \
       [expr $x + $options(x5) + $width]   $options(y1) \
         -width $options(-pageborderwidth) \
         -fill $options(-pageborderhilight) \
         -tags p$cnt

# draw normal border around right side of tab
    $widgets(canvas) create line \
       [expr $x + $options(x5) + $width] $options(y1) \
       [expr $x + $options(x6) + $width] $options(y2) \
       [expr $x + $options(x6) + $width] $options(y5) \
         -width $options(-pageborderwidth) \
         -tags p$cnt \
         -fill $options(-pagebordercolor)

    set x [expr $x + $options(x7) + $width]
    set options(p$cnt,x6) [expr $x - $options(-pageborderwidth)]

    if {![info exists widgets(f$cnt)]} {
      set widgets(f$cnt) [frame $widgets(canvas).f$cnt -bd 0]
    }

    $widgets(f$cnt) config -bg $options(-pagecolor)

    place $widgets(f$cnt) -x $options(x2) -y $options(y6) \
      -width $options(-pagewidth) -height $options(-pageheight)
    incr cnt
  }

  $widgets(canvas) create line \
     $options(x1) [expr $options(y5) - $options(-pageborderwidth)] \
     $options(x1) $options(y3) \
       -width $options(-pageborderwidth) \
       -fill $options(-pageborderhilight)

  $widgets(canvas) create line \
     $options(x1) $options(y3) \
     $options(x2) $options(y4) \
     $options(x3) $options(y4) \
     $options(x4) $options(y3) \
     $options(x4) $options(y6) \
     $options(x3) $options(y5) \
       -width $options(-pageborderwidth) \
       -fill $options(-pagebordercolor)

  $widgets(canvas) config \
    -width $options(-width) \
    -height $options(-height) \
    -bg $options(-background)
    
  set top $options(-top)
  set options(-top) -1
  raise.page $w $top
}


#
# This routine is called whenever the mouse-button is pressed over
# the notebook.  It determines if any page should be raised and raises
# that page.
#
proc ::notebook::Click {w x y} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  if {$y<$options(y1) || $y>$options(y6)} return
  set N [llength $options(-pages)]
  for {set i 0} {$i<$N} {incr i} {
    if {$x>=$options(p$i,x5) && $x<=$options(p$i,x6)} {
      ::notebook::raise.page $w $i
      break
    }
  }
}

#
# For internal use only.  This procedure raised the n-th page of
# the notebook
#
proc ::notebook::raise.page {w n} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  if {$n<0 || $n>=[llength $options(-pages)]} return
  set top $options(-top)
  if {$top >= 0 && $top < [llength $options(-pages)]} {
    $widgets(canvas) move p$top 0 $options(-pageborderwidth)
  }
  $widgets(canvas) move p$n 0 [expr -1*$options(-pageborderwidth)]
  $widgets(canvas) delete topline
  if {$n > 0} {
    $widgets(canvas) create line \
       $options(x1) $options(y5) \
       $options(p$n,x5) $options(y5) \
       $options(p$n,x5) [expr $options(y5) - $options(-pageborderwidth)] \
         -width $options(-pageborderwidth) \
         -fill $options(-pageborderhilight) \
         -tags topline
  }
  $widgets(canvas) create line \
    $options(p$n,x6) [expr $options(y5) - $options(-pageborderwidth)] \
    $options(p$n,x6) $options(y5) \
      -width $options(-pageborderwidth) \
      -fill $options(-pageborderhilight) \
      -tags topline
  $widgets(canvas) create line \
    $options(p$n,x6) $options(y5) \
    $options(x3) $options(y5) \
      -width $options(-pageborderwidth) \
      -fill $options(-pageborderhilight) \
      -tags topline
  set options(-top) $n
  ::raise $widgets(f$n)
}

#
# Change the page-specific configuration options for the notebook
#
proc ::notebook::pageconfig {w name args} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  set i [lsearch $options(-pages) $name]
  if {$i<0} return
  foreach {tag value} $args {
    switch -- $tag {
      -state {
        if {"$value"=="disabled"} {
          $widgets(canvas) itemconfig t$i -fg $options(fg,off)
        } else {
          $widgets(canvas) itemconfig t$i -fg $options(fg,on)
        }
      }
      -onexit {
        set options(p$i,onexit) $value
      }
    }
  }
}

#
# This procedure raises a notebook page given its name.  But first
# we check the "onexit" procedure for the current page (if any) and
# if it returns false, we don't allow the raise to proceed.
#
proc ::notebook::Raise {w name} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  set i [lsearch $options(-pages) $name]
  if {$i < 0} return
  if {[info exists options(p$i,onexit)]} {
    set onexit $options(p$i,onexit)
    if {"$onexit"!="" && [eval uplevel #0 $onexit]!=0} {
      raise.page $w $i
    }
  } else {
    raise.page $w $i
  }
}

#
# Return the frame associated with a given page of the notebook.
#
proc ::notebook::Nbframe {w name} {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  set i [lsearch $options(-pages) $name]
  if {$i>=0} {
    return $widgets(f$i)
  } else {
    return {}
  }
}

#
# Return the name of the visible frame
#
proc ::notebook::Top {w} {
  upvar ::notebook::${w}::options options
  return [lindex $options(-pages) $options(-top)]
}

#
# Try to resize the notebook to the next time we become idle.
#
proc ::notebook::ScheduleExpand w {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  if {[info exists options(expand)]} return
  set options(expand) 1
  after idle "::notebook::Expand ${w}"
}

#
# Resize the notebook to fit inside its containing widget.
#
proc ::notebook::Expand w {
  upvar ::notebook::${w}::widgets widgets
  upvar ::notebook::${w}::options options
  set wi [expr [winfo width $w]-($options(-pad)*2+4)]
  set hi [expr [winfo height $w]-($options(-pad)*2+36)]
  ::notebook::Configure $w -width $wi -height $hi
  catch {unset options(expand)}
}

# end of notebook.tcl

